@@ -9,9 +9,10 @@ from typing import Any, List, Optional
 import neuro_flow
 from neuro_flow.cli import batch, completion, images, live, storage
 from neuro_flow.parser import ConfigDir, find_workspace
-from neuro_flow.types import LocalPath
+from neuro_flow.types import LocalPath, TaskStatus
 
-from ..expr import MultiEvalError
+from ..batch_runner import BakeFailedError
+from ..expr import MultiError
 from .root import Root
 
 
@@ -197,11 +198,15 @@ def main(args: Optional[List[str]] = None) -> None:
         sys.exit(e.exit_code)
     except ClickExit as e:
         sys.exit(e.exit_code)
+    except BakeFailedError as e:
+        if e.status == TaskStatus.CANCELLED:
+            sys.exit(130)
+        sys.exit(1)
 
     except SystemExit:
         raise
 
-    except MultiEvalError as e:
+    except MultiError as e:
         for error in e.errors:
             LOG_ERROR(f"{error}")
         sys.exit(1)
