@@ -142,10 +142,27 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
         self, meta: JobMeta, suffix: Optional[str]
     ) -> AsyncIterator[JobDescription]:
         found = False
+        since = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
+            days=7
+        )
         if meta.multi and not suffix:
             async for job in self.client.jobs.list(
                 tags=meta.tags,
                 reverse=True,
+                statuses=(JobStatus.PENDING, JobStatus.RUNNING),
+            ):
+                found = True
+                yield job
+            async for job in self.client.jobs.list(
+                tags=meta.tags,
+                reverse=True,
+                statuses=(
+                    JobStatus.SUSPENDED,
+                    JobStatus.SUCCEEDED,
+                    JobStatus.FAILED,
+                    JobStatus.CANCELLED,
+                ),
+                since=since,
             ):
                 found = True
                 yield job
@@ -157,6 +174,22 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
                 tags=tags,
                 reverse=True,
                 limit=1,
+                statuses=(JobStatus.PENDING, JobStatus.RUNNING),
+            ):
+                found = True
+                yield job
+                return
+            async for job in self.client.jobs.list(
+                tags=tags,
+                reverse=True,
+                limit=1,
+                statuses=(
+                    JobStatus.SUSPENDED,
+                    JobStatus.SUCCEEDED,
+                    JobStatus.FAILED,
+                    JobStatus.CANCELLED,
+                ),
+                since=since,
             ):
                 found = True
                 yield job
@@ -307,6 +340,7 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
         dry_run: bool,
     ) -> None:
         """Run a named job"""
+        assert self._flow is not None
 
         meta_ctx = await self._ensure_meta(job_id, suffix, skip_check=True)
         is_multi = meta_ctx.multi
@@ -316,7 +350,9 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
                 "Additional job arguments are supported by multi-jobs only"
             )
 
-        if await self._try_attach_to_running(job_id, suffix, args, params):
+        if not dry_run and await self._try_attach_to_running(
+            job_id, suffix, args, params
+        ):
             return  # Attached to running job
 
         if not is_multi:
@@ -330,7 +366,23 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
         if job.title:
             run_args.append(f"--description={job.title}")
         if job.name:
-            run_args.append(f"--name={job.name}")
+            name = job.name
+        else:
+            first_part = self._flow.project.id.replace("_", "-").strip("-")
+            while "--" in first_part:
+                first_part = first_part.replace("--", "-")
+            second_part = job_id
+            if is_multi:
+                second_part += f"-{suffix}"
+            second_part = second_part.replace("_", "-").strip("-")
+            while "--" in second_part:
+                second_part = second_part.replace("--", "-")
+
+            first_part = first_part[: 40 - len(second_part) - 1]
+            name = first_part + "-" + second_part
+            while "--" in name:
+                name = name.replace("--", "-")
+        run_args.append(f"--name={name}")
         if job.preset is not None:
             run_args.append(f"--preset={job.preset}")
         if job.schedule_timeout is not None:
@@ -362,7 +414,6 @@ class LiveRunner(AsyncContextManager["LiveRunner"]):
             run_args.append(f"--port-forward={pf}")
         if job.pass_config:
             run_args.append(f"--pass-config")
-        assert self._flow is not None
         project_role = self._flow.project.role
         if project_role is not None:
             await self._create_project_role(project_role)
