@@ -43,6 +43,7 @@ from neuro_flow.types import ImageStatus, LocalPath
 from .config_loader import ConfigFile
 from .context import DepCtx, JobMeta
 from .types import FullID, TaskStatus
+from .utils import RetryConfig, async_retried, retry
 
 
 if sys.version_info < (3, 7):
@@ -143,9 +144,8 @@ class FinishedTask:
 @dataclasses.dataclass
 class BakeImage:
     id: str
-    prefix: Tuple[str, ...]
-    yaml_id: str
     ref: str
+    yaml_defs: Sequence[FullID]
     context_on_storage: Optional[URL]
     dockerfile_rel: Optional[str]
 
@@ -154,15 +154,16 @@ class BakeImage:
 
     @classmethod
     def from_primitive(cls, data: Mapping[str, Any]) -> "BakeImage":
+        context_on_storage_raw = data.get("context_on_storage")
+        context_on_storage = None
+        if context_on_storage_raw:
+            context_on_storage = URL(context_on_storage_raw)
         return cls(
             id=data["id"],
-            prefix=_id_from_json(data["prefix"]),
-            yaml_id=data["yaml_id"],
+            yaml_defs=[_id_from_json(it) for it in data["yaml_defs"]],
             ref=data["ref"],
-            context_on_storage=URL(data["context_on_storage"])
-            if "context_on_storage" in data
-            else None,
-            dockerfile_rel=data["dockerfile_rel"] if "dockerfile_rel" in data else None,
+            context_on_storage=context_on_storage,
+            dockerfile_rel=data.get("dockerfile_rel"),
             status=ImageStatus(data["status"]),
             builder_job_id=data["builder_job_id"],
         )
@@ -267,7 +268,9 @@ class Storage(abc.ABC):
         pass
 
     @abc.abstractmethod
-    async def find_attempt(self, bake: Bake, attempt_no: int = -1) -> Attempt:
+    async def find_attempt(
+        self, bake: Bake, attempt_no: int = -1, force_no_cache: bool = False
+    ) -> Attempt:
         pass
 
     @abc.abstractmethod
@@ -438,15 +441,16 @@ class Storage(abc.ABC):
         pass
 
     @abc.abstractmethod
-    async def clear_cache(self, project: str, batch: Optional[str] = None) -> None:
+    async def clear_cache(
+        self, project: str, batch: Optional[str] = None, task_id: Optional[str] = None
+    ) -> None:
         pass
 
     @abc.abstractmethod
     async def create_bake_image(
         self,
         bake: Bake,
-        prefix: Tuple[str, ...],
-        yaml_id: str,
+        yaml_defs: Sequence[FullID],
         ref: str,
         context_on_storage: Optional[URL],
         dockerfile_rel: Optional[str],
@@ -786,7 +790,9 @@ class FSStorage(Storage):
         await self._write_json(attempt_uri / f"{pre}.init.json", _attempt_to_json(ret))
         return ret
 
-    async def find_attempt(self, bake: Bake, attempt_no: int = -1) -> Attempt:
+    async def find_attempt(
+        self, bake: Bake, attempt_no: int = -1, force_no_cache: bool = False
+    ) -> Attempt:
         bake_uri = _mk_bake_uri(self._fs, bake)
         if attempt_no == -1:
             files = set()
@@ -991,14 +997,18 @@ class FSStorage(Storage):
             await self._fs.mkdir(url.parent, parents=True)
             await self._write_json(url, data, overwrite=True)
 
-    async def clear_cache(self, project: str, batch: Optional[str] = None) -> None:
-        await self._fs.rm(_mk_cache_uri2(self._fs, project, batch), recursive=True)
+    async def clear_cache(
+        self, project: str, batch: Optional[str] = None, task_id: Optional[str] = None
+    ) -> None:
+        url = _mk_cache_uri2(self._fs, project, batch)
+        if task_id:
+            url = url / f"{task_id}.json"
+        await self._fs.rm(url, recursive=True)
 
     async def create_bake_image(
         self,
         bake: Bake,
-        prefix: Tuple[str, ...],
-        yaml_id: str,
+        yaml_defs: Sequence[FullID],
         ref: str,
         context_on_storage: Optional[URL],
         dockerfile_rel: Optional[str],
@@ -1283,7 +1293,7 @@ class APIStorage(Storage):
             else:
                 for key, action_meta in configs_meta["action_configs"].items():
                     if filename == action_meta["storage_filename"]:
-                        filename == action_meta["real_name"]
+                        filename = action_meta["real_name"]
                         real_meta["action_config_ids"][key] = await self._write_config(
                             bake_data["id"], filename, config.content
                         )
@@ -1293,6 +1303,7 @@ class APIStorage(Storage):
         await self._create_attempt(bake, 1, real_meta, when)
         return bake
 
+    @async_retried("Failed to fetch bake")
     async def fetch_bake(
         self, project: str, batch: str, when: datetime.datetime, suffix: str
     ) -> Bake:
@@ -1313,12 +1324,14 @@ class APIStorage(Storage):
 
         raise ResourceNotFound
 
+    @async_retried("Failed to fetch bake by id")
     async def fetch_bake_by_id(self, project: str, bake_id: str) -> Bake:
         batch, whenstr, suffix = bake_id.split("_")
         when = datetime.datetime.fromisoformat(whenstr)
 
         return await self.fetch_bake(project, batch, when, "")
 
+    @async_retried("Failed to fetch bakes by name")
     async def fetch_bake_by_name(self, project: str, name: str) -> Bake:
         prj = await self._get_project(project)
         url = self._base_url / "api/v1/flow/bakes/by_name"
@@ -1392,6 +1405,7 @@ class APIStorage(Storage):
             payload = await resp.json()
             return cast(Dict[str, Any], payload)
 
+    @async_retried("Failed to fetch configs")
     async def fetch_config(self, bake: Bake, filename: str) -> str:
         # filename is config id actually
         ret = await self._get_config(filename)
@@ -1463,22 +1477,23 @@ class APIStorage(Storage):
         )
 
     async def _find_attempt_data(
-        self, bake: Bake, attempt_no: int = -1
+        self, bake: Bake, attempt_no: int = -1, force_no_cache: bool = False
     ) -> Dict[str, Any]:
         bake_data = await self._find_bake_data(bake)
         last_attempts = []
 
-        for attempt_data in self._attempts_cache.values():
-            if attempt_data["bake_id"] == bake_data["id"]:
-                if attempt_no != -1:
-                    if attempt_data["number"] == attempt_no:
-                        return attempt_data
-                else:
-                    last_attempts.append(attempt_data)
+        if not force_no_cache:
+            for attempt_data in self._attempts_cache.values():
+                if attempt_data["bake_id"] == bake_data["id"]:
+                    if attempt_no != -1:
+                        if attempt_data["number"] == attempt_no:
+                            return attempt_data
+                    else:
+                        last_attempts.append(attempt_data)
 
-        if last_attempts:
-            last_attempts.sort(key=itemgetter("number"))
-            return last_attempts[-1]
+            if last_attempts:
+                last_attempts.sort(key=itemgetter("number"))
+                return last_attempts[-1]
 
         assert attempt_no == -1 or 0 < attempt_no < 99
         url = self._base_url / "api/v1/flow/attempts/by_number"
@@ -1496,10 +1511,14 @@ class APIStorage(Storage):
             self._attempts_cache[attempt_data["id"]] = attempt_data
             return attempt_data
 
-    async def find_attempt(self, bake: Bake, attempt_no: int = -1) -> Attempt:
-        attempt_data = await self._find_attempt_data(bake, attempt_no)
+    @async_retried("Failed to find attempt")
+    async def find_attempt(
+        self, bake: Bake, attempt_no: int = -1, force_no_cache: bool = False
+    ) -> Attempt:
+        attempt_data = await self._find_attempt_data(bake, attempt_no, force_no_cache)
         return _attempt_from_api_json(bake, attempt_data)
 
+    @async_retried("Failed to fetch attempt")
     async def fetch_attempt(
         self, attempt: Attempt
     ) -> Tuple[Dict[FullID, StartedTask], Dict[FullID, FinishedTask]]:
@@ -1729,7 +1748,7 @@ class APIStorage(Storage):
         st = StartedTask(
             attempt=attempt,
             id=task_id,
-            raw_id="",
+            raw_id=payload.get("raw_id", ""),
             when=created_at,
             created_at=created_at,
         )
@@ -1737,7 +1756,7 @@ class APIStorage(Storage):
         ft = FinishedTask(
             attempt=attempt,
             id=task_id,
-            raw_id="",
+            raw_id=payload.get("raw_id", ""),
             when=created_at,
             status=TaskStatus.CACHED,
             created_at=created_at,
@@ -1766,6 +1785,7 @@ class APIStorage(Storage):
                 "task_id": _id_to_json(ft.id),
                 "batch": attempt.bake.batch,
                 "key": caching_key,
+                "raw_id": ft.raw_id,
                 "outputs": ft.outputs,
                 "state": ft.state,
             },
@@ -1796,28 +1816,37 @@ class APIStorage(Storage):
             await self._fs.mkdir(url.parent, parents=True)
             await self._write_json(url, data, overwrite=True)
 
-    async def clear_cache(self, project: str, batch: Optional[str] = None) -> None:
+    async def clear_cache(
+        self, project: str, batch: Optional[str] = None, task_id: Optional[str] = None
+    ) -> None:
         prj = await self._get_project(project)
 
+        params = {
+            "project_id": prj.id,
+        }
+        if batch:
+            params["batch"] = batch
+        if task_id:
+            params["task_id"] = task_id
+
         auth = await self._config._api_auth()
         async with self._core.request(
             "DELETE",
             self._base_url / "api/v1/flow/cache_entries",
-            params={
-                "project_id": prj.id,
-                "batch": batch,
-            },
+            params=params,
             auth=auth,
         ) as resp:
-            await resp.json()
+            resp.raise_for_status()
 
-        await self._fs.rm(_mk_cache_uri2(self._fs, project, batch), recursive=True)
+        url = _mk_cache_uri2(self._fs, project, batch)
+        if task_id:
+            url = url / f"{task_id}.json"
+        await self._fs.rm(url, recursive=True)
 
     async def create_bake_image(
         self,
         bake: Bake,
-        prefix: Tuple[str, ...],
-        yaml_id: str,
+        yaml_defs: Sequence[FullID],
         ref: str,
         context_on_storage: Optional[URL],
         dockerfile_rel: Optional[str],
@@ -1830,8 +1859,7 @@ class APIStorage(Storage):
         image_data = dict(
             {
                 "bake_id": bake_data["id"],
-                "yaml_id": yaml_id,
-                "prefix": _id_to_json(prefix),
+                "yaml_defs": [_id_to_json(it) for it in yaml_defs],
                 "ref": ref,
                 "context_on_storage": str(context_on_storage)
                 if context_on_storage
@@ -1872,6 +1900,7 @@ class APIStorage(Storage):
                 image_data = json.loads(line)
                 yield BakeImage.from_primitive(image_data)
 
+    @async_retried("Failed to get bake image")
     async def get_bake_image(self, bake: Bake, ref: str) -> BakeImage:
         bake_data = await self._find_bake_data(bake)
 
@@ -1915,6 +1944,7 @@ class APIStorage(Storage):
         ) as resp:
             return BakeImage.from_primitive(await resp.json())
 
+    @async_retried("Failed to write file")  # File write is idempotent => ok to retry
     async def _write_file(
         self, url: URL, body: str, *, overwrite: bool = False
     ) -> None:
@@ -2135,3 +2165,220 @@ def _find_finished_at(statuses: List[_StatusItem]) -> datetime.datetime:
             return item["created_at"]
     else:
         raise ValueError(f"Task is not finished, the history is {statuses}")
+
+
+class RetryReadStorage(Storage, RetryConfig):
+    # Storage implementation that retries reads.
+    # Intentded to make batch executor more stable
+
+    def __init__(self, storage: Storage) -> None:
+        super().__init__()
+        self._storage = storage
+
+    async def close(self) -> None:
+        await self._storage.close()
+
+    async def ensure_project(
+        self, name: str, owner: Optional[str] = None, cluster: Optional[str] = None
+    ) -> Project:
+        return await self._storage.ensure_project(
+            name=name, owner=owner, cluster=cluster
+        )
+
+    async def write_live(self, project: str, jobs: Iterable[JobMeta]) -> Live:
+        raise NotImplementedError
+
+    @retry
+    async def _list_bakes(
+        self,
+        project: str,
+        tags: Optional[AbstractSet[str]] = None,
+        since: Optional[datetime.datetime] = None,
+        until: Optional[datetime.datetime] = None,
+        recent_first: bool = False,
+    ) -> List[Bake]:
+        bakes = []
+        async for bake in self._storage.list_bakes(
+            project=project,
+            tags=tags,
+            since=since,
+            until=until,
+            recent_first=recent_first,
+        ):
+            bakes.append(bake)
+        return bakes
+
+    async def list_bakes(
+        self,
+        project: str,
+        tags: Optional[AbstractSet[str]] = None,
+        since: Optional[datetime.datetime] = None,
+        until: Optional[datetime.datetime] = None,
+        recent_first: bool = False,
+    ) -> AsyncIterator[Bake]:
+        for bake in await self._list_bakes(
+            project=project,
+            tags=tags,
+            since=since,
+            until=until,
+            recent_first=recent_first,
+        ):
+            yield bake
+
+    async def create_bake(
+        self,
+        project: str,
+        batch: str,
+        configs_meta: Mapping[str, Any],
+        configs: Sequence[ConfigFile],
+        graphs: Mapping[FullID, Mapping[FullID, AbstractSet[FullID]]],
+        params: Optional[Mapping[str, str]],
+        name: Optional[str],
+        tags: Sequence[str] = (),
+    ) -> Bake:
+        return await self._storage.create_bake(
+            project=project,
+            batch=batch,
+            configs_meta=configs_meta,
+            configs=configs,
+            graphs=graphs,
+            params=params,
+            name=name,
+            tags=tags,
+        )
+
+    @retry
+    async def fetch_bake(
+        self, project: str, batch: str, when: datetime.datetime, suffix: str
+    ) -> Bake:
+        return await self._storage.fetch_bake(
+            project=project, batch=batch, when=when, suffix=suffix
+        )
+
+    @retry
+    async def fetch_bake_by_id(self, project: str, bake_id: str) -> Bake:
+        return await self._storage.fetch_bake_by_id(project=project, bake_id=bake_id)
+
+    @retry
+    async def fetch_bake_by_name(self, project: str, bake_name: str) -> Bake:
+        return await self._storage.fetch_bake_by_name(
+            project=project, bake_name=bake_name
+        )
+
+    @retry
+    async def fetch_configs_meta(self, bake: Bake) -> Mapping[str, Any]:
+        return await self._storage.fetch_configs_meta(bake=bake)
+
+    @retry
+    async def fetch_config(self, bake: Bake, filename: str) -> str:
+        return await self._storage.fetch_config(bake=bake, filename=filename)
+
+    @retry
+    async def create_attempt(self, bake: Bake, attempt_no: int) -> Attempt:
+        return await self._storage.create_attempt(bake=bake, attempt_no=attempt_no)
+
+    @retry
+    async def find_attempt(
+        self, bake: Bake, attempt_no: int = -1, force_no_cache: bool = False
+    ) -> Attempt:
+        return await self._storage.find_attempt(
+            bake=bake, attempt_no=attempt_no, force_no_cache=force_no_cache
+        )
+
+    @retry
+    async def fetch_attempt(
+        self, attempt: Attempt
+    ) -> Tuple[Dict[FullID, StartedTask], Dict[FullID, FinishedTask]]:
+        return await self._storage.fetch_attempt(attempt=attempt)
+
+    async def store_executor_id(self, attempt: Attempt, executor_id: str) -> None:
+        await self._storage.store_executor_id(attempt=attempt, executor_id=executor_id)
+
+    async def finish_attempt(self, attempt: Attempt, result: TaskStatus) -> None:
+        await self._storage.finish_attempt(attempt=attempt, result=result)
+
+    async def write_start(
+        self,
+        task: StartedTask,
+    ) -> None:
+        await self._storage.write_start(task)
+
+    async def write_finish(
+        self,
+        task: FinishedTask,
+    ) -> None:
+        await self._storage.write_finish(task)
+
+    @retry
+    async def check_cache(
+        self,
+        attempt: Attempt,
+        task_id: FullID,
+        caching_key: str,
+        life_span: datetime.timedelta,
+    ) -> Optional[FinishedTask]:
+        return await self._storage.check_cache(
+            attempt=attempt,
+            task_id=task_id,
+            caching_key=caching_key,
+            life_span=life_span,
+        )
+
+    async def write_cache(
+        self,
+        attempt: Attempt,
+        ft: FinishedTask,
+        caching_key: str,
+    ) -> None:
+        await self._storage.write_cache(
+            attempt=attempt,
+            ft=ft,
+            caching_key=caching_key,
+        )
+
+    async def clear_cache(
+        self, project: str, batch: Optional[str] = None, task_id: Optional[str] = None
+    ) -> None:
+        await self._storage.clear_cache(project=project, batch=batch, task_id=task_id)
+
+    async def create_bake_image(
+        self,
+        bake: Bake,
+        yaml_defs: Sequence[FullID],
+        ref: str,
+        context_on_storage: Optional[URL],
+        dockerfile_rel: Optional[str],
+    ) -> BakeImage:
+        return await self._storage.create_bake_image(
+            bake=bake,
+            yaml_defs=yaml_defs,
+            ref=ref,
+            context_on_storage=context_on_storage,
+            dockerfile_rel=dockerfile_rel,
+        )
+
+    @retry
+    async def _list_bake_images(self, bake: Bake) -> List[BakeImage]:
+        images = []
+        async for image in self._storage.list_bake_images(bake=bake):
+            images.append(image)
+        return images
+
+    async def list_bake_images(self, bake: Bake) -> AsyncIterator[BakeImage]:
+        for image in await self._list_bake_images(bake=bake):
+            yield image
+
+    @retry
+    async def get_bake_image(self, bake: Bake, ref: str) -> BakeImage:
+        return await self._storage.get_bake_image(bake=bake, ref=ref)
+
+    async def update_bake_image(
+        self,
+        bake: Bake,
+        ref: str,
+        status: Union[ImageStatus, Type[_Unset]] = _Unset,
+        builder_job_id: Union[Optional[str], Type[_Unset]] = _Unset,
+    ) -> BakeImage:
+        return await self._storage.update_bake_image(
+            bake=bake, ref=ref, status=status, builder_job_id=builder_job_id
+        )
